@@ -5,10 +5,13 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.78  11-Feb-2000  miquels@cistron.nl
- *
+ * Version:	@(#)init.c  2.83  05-Sep-2001  miquels@cistron.nl
+ */
+#define VERSION "2.83"
+#define DATE    "01-Oct-2001"
+/*
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-2000 Miquel van Smoorenburg.
+ *		Copyright 1991-2001 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -49,14 +52,11 @@
 #ifdef __i386__
 #  if (__GLIBC__ >= 2)
      /* GNU libc 2.x */
-#    if (__GLIBC_MINOR__ == 0)
+#    define STACK_DEBUG 1
+#    if (__GLIBC__ == 2 && __GLIBC_MINOR__ == 0)
        /* Only glibc 2.0 needs this */
 #      include <sigcontext.h>
 #    endif
-#  else
-     /* Linux libc5 */
-#    include <asm/sigcontext.h>
-#    define sigcontext sigcontext_struct
 #  endif
 #endif
 
@@ -80,9 +80,9 @@
 		} while(0)
 
 /* Version information */
-char *Version = "@(#) init 2.78  11-Feb-2000  miquels@cistron.nl";
-char *bootmsg = "version 2.78 %s";
-#define E_VERSION "INIT_VERSION=sysvinit-2.78"
+char *Version = "@(#) init " VERSION "  " DATE "  miquels@cistron.nl";
+char *bootmsg = "version " VERSION " %s";
+#define E_VERSION "INIT_VERSION=sysvinit-" VERSION
 
 CHILD *family = NULL;		/* The linked list of all entries */
 CHILD *newFamily = NULL;	/* The list after inittab re-read */
@@ -382,7 +382,7 @@ static CHILD *get_record(FILE *f)
 				if (strcmp(actions[i].name, s) == 0)
 					break;
 			}
-			p->action = actions[i].act ? : OFF;
+			p->action = actions[i].act ? actions[i].act : OFF;
 			break;
 		default:
 			free(p);
@@ -416,9 +416,7 @@ int receive_state(int fd)
 }
 
 /*
- *	Set the process title. We do not check for overflow of
- *	the stack space since we know there is plenty for
- *	our needs and we'll never use more than 10 bytes anyway.
+ *	Set the process title.
  */
 int setproctitle(char *fmt, ...)
 {
@@ -429,7 +427,7 @@ int setproctitle(char *fmt, ...)
 	buf[0] = 0;
 
 	va_start(ap, fmt);
-	len = vsprintf(buf, fmt, ap);
+	len = vsnprintf(buf, sizeof(buf), fmt, ap);
 	va_end(ap);
 
 	memset(argv0, 0, maxproclen + 1);
@@ -439,7 +437,7 @@ int setproctitle(char *fmt, ...)
 }
 
 /*
- *	Reset console tries.
+ *	Set console_dev to a working console.
  */
 void console_init(void)
 {
@@ -592,7 +590,7 @@ int coredump(void)
  *	If we have the info, print where it occured.
  *	Then sleep 30 seconds and try to continue.
  */
-#ifdef __i386__
+#ifdef STACK_DEBUG
 void segv_handler(int sig, struct sigcontext ctx)
 {
 	char	*p = "";
@@ -684,7 +682,7 @@ void set_term(int how)
 		(void) tcgetattr(fd, &tty);
 
 		tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
-		tty.c_cflag |= HUPCL|CLOCAL;
+		tty.c_cflag |= HUPCL|CLOCAL|CREAD;
 
 		tty.c_cc[VINTR]  = 3;	/* ctrl('c') */
 		tty.c_cc[VQUIT]  = 28;	/* ctrl('\\') */
@@ -728,6 +726,9 @@ void print(char *s)
 /*
  *	Log something to a logfile and the console.
  */
+#ifdef __GNUC__
+__attribute__ ((format (printf, 2, 3)))
+#endif
 void log(int loglevel, char *s, ...)
 {
 	va_list va_alist;
@@ -742,8 +743,8 @@ void log(int loglevel, char *s, ...)
 		 *	Re-etablish connection with syslogd every time.
 		 */
 		openlog("init", 0, LOG_DAEMON);
-		syslog(LOG_INFO, buf);
-		/* closelog(); NOT needed with recent libc's. */
+		syslog(LOG_INFO, "%s", buf);
+		closelog();
 	}
 
 	/*
@@ -856,6 +857,7 @@ int spawn(CHILD *ch, int *res)
   } else {
 	/* Split up command line arguments */
   	strncpy(buf, proc, sizeof(buf) - 1);
+	buf[sizeof(buf) - 1] = 0;
   	ptr = buf;
   	for(f = 1; f < 15; f++) {
   		/* Skip white space */
@@ -909,8 +911,8 @@ int spawn(CHILD *ch, int *res)
 
 		/*
 		 *	In sysinit, boot, bootwait or single user mode:
-		 *	any wait-type subprocess gets the console
-		 *	as its controlling tty (if not in use yet).
+		 *	for any wait-type subprocess we _force_ the console
+		 *	to be its controlling tty.
 		 */
   		if (strchr("*#sS", runlevel) && ch->flags & WAITING) {
 			/*
@@ -920,7 +922,8 @@ int spawn(CHILD *ch, int *res)
 			 */
 			setsid();
 			if ((f = console_open(O_RDWR|O_NOCTTY)) >= 0) {
-				(void)ioctl(f, TIOCSCTTY, 0);
+				/* Take over controlling tty by force */
+				(void)ioctl(f, TIOCSCTTY, 1);
   				dup(f);
   				dup(f);
 			}
@@ -931,22 +934,28 @@ int spawn(CHILD *ch, int *res)
 			if (pid > 0) {
 				/*
 				 *	Ignore keyboard signals etc.
+				 *	Then wait for child to exit.
 				 */
 				SETSIG(sa, SIGINT, SIG_IGN, SA_RESTART);
 				SETSIG(sa, SIGTSTP, SIG_IGN, SA_RESTART);
 				SETSIG(sa, SIGQUIT, SIG_IGN, SA_RESTART);
 				SETSIG(sa, SIGCHLD, SIG_DFL, SA_RESTART);
 
-				while(waitpid(pid, &st, 0) != pid)
-					;
-				pgrp = 0;
+				while ((f = waitpid(pid, &st, 0)) != pid)
+					if (f < 0 && errno == ECHILD)
+						break;
+
+				/*
+				 *	Small optimization. See if stealing
+				 *	controlling tty back is needed.
+				 */
 				pgrp = tcgetpgrp(f);
 				if (pgrp != getpid())
 					exit(0);
+
 				/*
-				 *	De-attach session from controlling tty.
-				 *	Or actually, the other way around.
-				 *	This _without_ SIGHUP etc.
+				 *	Steal controlling tty away. We do
+				 *	this with a temporary process.
 				 */
 				if ((pid = fork()) < 0) {
   					log(L_VB, "cannot fork");
@@ -957,13 +966,17 @@ int spawn(CHILD *ch, int *res)
 					(void)ioctl(f, TIOCSCTTY, 1);
 					exit(0);
 				}
-				while(waitpid(pid, &st, 0) != pid)
-					;
+				while((f = waitpid(pid, &st, 0)) != pid)
+					if (f < 0 && errno == ECHILD)
+						break;
 				exit(0);
 			}
+
 			/* Set ioctl settings to default ones */
 			set_term(0);
+
   		} else {
+			setsid();
 			if ((f = console_open(O_RDWR|O_NOCTTY)) < 0) {
 				log(L_VB, "open(%s): %s", console_dev,
 					strerror(errno));
@@ -971,7 +984,6 @@ int spawn(CHILD *ch, int *res)
 			}
 			dup(f);
 			dup(f);
-			setsid();
 		}
 
   		/* Reset all the signals */
@@ -1003,7 +1015,7 @@ int spawn(CHILD *ch, int *res)
 #endif
 
 	if (pid == -1) {
-		log(L_VB, "cannot fork, retry..", NULL, NULL);
+		log(L_VB, "cannot fork, retry..");
 		do_sleep(5);
 		continue;
 	}
@@ -2451,6 +2463,8 @@ int main(int argc, char *argv[])
   	p = argv[0];
   	
 
+  umask(022);
+
   /*
    *	Is this telinit or init ?
    */
@@ -2524,8 +2538,6 @@ int main(int argc, char *argv[])
   if (argc - optind != 1 || strlen(argv[optind]) != 1) Usage(p);
   if (!strchr("0123456789SsQqAaBbCcUu", argv[optind][0])) Usage(p);
 
-  umask(022);
-
   /* Open the fifo and write a command. */
   memset(&request, 0, sizeof(request));
   request.magic     = INIT_MAGIC;
